package org.basex.query.value.item;

import static org.basex.query.QueryError.*;

import org.basex.query.*;
import org.basex.query.func.*;
import org.basex.query.util.collation.*;
import org.basex.query.util.list.*;
import org.basex.query.value.map.*;
import org.basex.query.value.node.*;
import org.basex.query.value.type.*;
import org.basex.util.*;

/**
 * Abstract super class for function items.
 * This class is inherited by {@link XQMap}, {@link Array}, and {@link FuncItem}.
 *
 * @author BaseX Team 2005-19, BSD License
 * @author Leo Woerteler
 */
public abstract class FItem extends Item implements XQFunction {
  /** Annotations. */
  final AnnList anns;

  /**
   * Constructor.
   * @param type type
   * @param anns this function item's annotations
   */
  protected FItem(final FuncType type, final AnnList anns) {
    super(type);
    this.anns = anns;
  }

  @Override
  public final AnnList annotations() {
    return anns;
  }

  @Override
  public final byte[] string(final InputInfo info) throws QueryException {
    throw FIATOM_X.get(info, type);
  }

  @Override
  public final boolean eq(final Item item, final Collation coll, final StaticContext sc,
      final InputInfo info) throws QueryException {
    throw FIATOM_X.get(info, type);
  }

  @Override
  public final boolean sameKey(final Item item, final InputInfo info) {
    return false;
  }

  /**
   * Coerces this function item to the given function type.
   * @param ft function type
   * @param qc query context
   * @param info input info
   * @param optimize optimize resulting item
   * @return coerced item
   * @throws QueryException query exception
   */
  public abstract FItem coerceTo(FuncType ft, QueryContext qc, InputInfo info, boolean optimize)
      throws QueryException;

  /**
   * Performs a deep comparison of two items.
   * @param item item to be compared
   * @param info input info
   * @param coll collation (can be {@code null})
   * @return result of check
   * @throws QueryException query exception
   */
  public abstract boolean deep(Item item, InputInfo info, Collation coll) throws QueryException;

  @Override
  public abstract void plan(FElem root);
}
